import ddf.minim.*;
import ddf.minim.signals.*;
import ddf.minim.analysis.*;
import ddf.minim.effects.*;


import ddf.minim.*;

PImage startScreen;
PImage finalWinScreen;

PImage grassBackground;
PImage lavaImage;
PImage shrubImage;
PImage blackSheepImage;
PImage sheepImage;
PImage leftOnlySheepImage;
PImage reverseSheepImage;
PImage flowersImage;

boolean showStartScreen = true;
boolean showLevelScreen = false;
boolean showWinScreen = false;
boolean showFinalWinScreen = false;


int transitionDuration = 10;
int currentTransitionFrame = 0;
boolean transitioning = false;

int level = 1;
String levelTitle;
String levelDescription;

int goalX;
int goalY;


int gridWidthInSquares = 10;
int gridHeightInSquares = 10;

ArrayList sheeps = new ArrayList();
ArrayList shrubs = new ArrayList();
ArrayList lavas = new ArrayList();
ArrayList gates = new ArrayList();
ArrayList buttons = new ArrayList();

AudioPlayer music;
AudioSample step1;
AudioSample step2;
AudioSample step3;
AudioSample step4;
AudioSample winSound;
AudioSample dieSound;
Minim minim;

void setup() 
{
  size(500, 500);
  minim = new Minim(this);
  step1 = minim.loadSample("bell1.wav", 2048);
  step2 = minim.loadSample("bell2.wav", 2048);
  step3 = minim.loadSample("bell3.wav", 2048);
  step4 = minim.loadSample("bell4.wav", 2048);
  dieSound = minim.loadSample("step2.wav", 2048);
  winSound = minim.loadSample("baah.wav", 2048);
  try {
    music = minim.loadFile("fiddle4.wav", 2048);
    music.setLoopPoints(0, 138759);
  }
  catch(Exception e) {
  }

  if (!online) {
    frame.setTitle("Black Sheep");
  }

  ellipseMode(RADIUS);

  frameRate(30);
  smooth();

  startScreen = loadImage("startScreen.jpg");
  finalWinScreen = loadImage("finalWin.jpg");
  grassBackground = loadImage("grassBackground.jpg");
  lavaImage = loadImage("lava.jpg");
  shrubImage = loadImage("shrub5.png");
  sheepImage = loadImage("sheep8.png");
  blackSheepImage = loadImage("sheep7.png");
  reverseSheepImage = loadImage("sheep9.png");
  leftOnlySheepImage = loadImage("sheep4.png");
  flowersImage = loadImage("flowers.png");
  restart();

  winSound.trigger();
  music.loop();
}

void stop()
{
  // always close Minim audio classes when you are done with them
  if (music != null) {
    music.close();
  }
  winSound.close();
  dieSound.close();
  step1.close();
  step2.close();
  step3.close();
  step4.close();
  minim.stop();

  super.stop();
}


public void restart() {
 
  transitioning = false;

  sheeps.clear();
  shrubs.clear();
  lavas.clear();
  buttons.clear();
  gates.clear();

  if (level == 1) {

    gridWidthInSquares = 10;
    gridHeightInSquares = 10;

    goalX = 9;
    goalY = 8;


    levelTitle = "Baah World";
    levelDescription = "Use the arrow keys to escape the pasture!";

    shrubs.add(new Shrub(1, 2));
    shrubs.add(new Shrub(1, 3));

    for (int r = 2; r < 10; r++) {
      shrubs.add(new Shrub(2, r));
    }

    shrubs.add(new Shrub(3, 2));
    shrubs.add(new Shrub(3, 3));



    shrubs.add(new Shrub(4, 6));
    shrubs.add(new Shrub(4, 7));

    for (int r = 0; r < 8; r++) {
      shrubs.add(new Shrub(5, r));
    }

    shrubs.add(new Shrub(6, 6));
    shrubs.add(new Shrub(6, 7));

    for (int r = 3; r < 10; r++) {
      shrubs.add(new Shrub(8, r));
    }

    shrubs.add(new Shrub(7, 3));

    sheeps.add(new Sheep(1, 8, false));
  }
  else if (level == 2) {

    gridWidthInSquares = 10;
    gridHeightInSquares = 10;

    goalX = 9;
    goalY = 5;


    levelTitle = "Hello Sheep";
    levelDescription = "Most sheep follow the herd.";



    for (int c = 0; c < 10; c++) {
      shrubs.add(new Shrub(c, 3));
      shrubs.add(new Shrub(c, 6));
    }


    sheeps.add(new Sheep(0, 4, false));
    sheeps.add(new Sheep(0, 0, true));
    sheeps.add(new Sheep(0, 9, true));
  }
  else if (level == 3) { 
    gridWidthInSquares = 10;
    gridHeightInSquares = 10;

    goalX = 9;
    goalY = 5;


    levelTitle = "Predictably Opposing";
    levelDescription = "Some sheep move in different ways.";



    for (int c = 0; c < 10; c++) {
      shrubs.add(new Shrub(c, 3));
      shrubs.add(new Shrub(c, 6));
    }


    sheeps.add(new Sheep(1, 4, false));
    sheeps.add(new Sheep(0, 1, true));
    sheeps.add(new Sheep(0, 8, true));
    sheeps.add(new ReverseSheep(8, 4, true));
    sheeps.add(new ReverseSheep(9, 1, true));
    sheeps.add(new ReverseSheep(9, 8, true));
  }
    else if (level == 4) { 
    gridWidthInSquares = 10;
    gridHeightInSquares = 10;

    goalX = 9;
    goalY = 0;


    levelTitle = "Environmentally Unfriendly";
    levelDescription = "Use your environment!";




    for (int c = 0; c < 10; c++) {
      if (c != 4) {
        shrubs.add(new Shrub(c, 5));
      }
    }
    for (int r = 0; r < 10; r++) {
      if (r < 3 || r > 6) {
        shrubs.add(new Shrub(4, r));
      }
    }



    sheeps.add(new Sheep(0, 9, false));
    sheeps.add(new Sheep(1, 9, true));
    sheeps.add(new Sheep(2, 9, true));


    sheeps.add(new Sheep(2, 2, true));

    sheeps.add(new ReverseSheep(4, 5, true));
    sheeps.add(new ReverseSheep(5, 6, true));
    sheeps.add(new ReverseSheep(6, 3, true));
  }
    else if (level == 5) {

    gridWidthInSquares = 10;
    gridHeightInSquares = 10;

    goalX = 9;
    goalY = 9;


    levelTitle = "Whoops";
    levelDescription = "Press R to restart";
    sheeps.add(new Sheep(0, 0, false));

    sheeps.add(new Sheep(1, 0, true));

    sheeps.add(new LeftOnlySheep(5, 0, true));
    sheeps.add(new LeftOnlySheep(1, 7, true));

    shrubs.add(new Shrub(1, 1));
    shrubs.add(new Shrub(1, 2));
    shrubs.add(new Shrub(1, 3));
    shrubs.add(new Shrub(1, 4));
    shrubs.add(new Shrub(1, 5));
    shrubs.add(new Shrub(1, 6));
    shrubs.add(new Shrub(2, 7));

    shrubs.add(new Shrub(3, 7));
    shrubs.add(new Shrub(4, 7));
    shrubs.add(new Shrub(5, 7));
    shrubs.add(new Shrub(6, 7));
    shrubs.add(new Shrub(7, 7));
  }
  else if (level == 6) { 
    gridWidthInSquares = 10;
    gridHeightInSquares = 10;

    goalX = 9;
    goalY = 9;


    levelTitle = "Butting Heads";
    levelDescription = "Use the herd against itself!";


    shrubs.add(new Shrub(0, 7));
    shrubs.add(new Shrub(1, 7));
    shrubs.add(new Shrub(2, 7));
    shrubs.add(new Shrub(2, 8));
    shrubs.add(new Shrub(2, 9));

    shrubs.add(new Shrub(7, 0));
    shrubs.add(new Shrub(7, 1));
    shrubs.add(new Shrub(7, 2));
    shrubs.add(new Shrub(8, 2));
    shrubs.add(new Shrub(9, 2));



    sheeps.add(new Sheep(0, 0, false));
    sheeps.add(new Sheep(2, 0, true));
    sheeps.add(new Sheep(2, 1, true));
    sheeps.add(new Sheep(2, 2, true));
    sheeps.add(new Sheep(1, 2, true));
    sheeps.add(new Sheep(0, 2, true));


    sheeps.add(new ReverseSheep(9, 7, true));
    sheeps.add(new ReverseSheep(8, 7, true));
    sheeps.add(new ReverseSheep(7, 7, true));
    sheeps.add(new ReverseSheep(7, 8, true));
    sheeps.add(new ReverseSheep(7, 9, true));
  }



  else if (level == 7) {

    gridWidthInSquares = 10;
    gridHeightInSquares = 10;

    goalX = 9;
    goalY = 9;


    levelTitle = "35 Slider";
    levelDescription = "Press R to restart";

    sheeps.add(new Sheep(0, 0, false));




    for (int c = 0; c < 10; c++) {

      if (c != 0) {
        sheeps.add(new Sheep(c, 0, true));
      }
      if (c != 9) {
        sheeps.add(new Sheep(c, 9, true));
      }
    }

    for (int r = 1; r < 9; r++) {

      if (r != 0) {
        sheeps.add(new Sheep(0, r, true));
      }
      if (r != 9) {
        sheeps.add(new Sheep(9, r, true));
      }
    }

    for (int r = 1; r < 9; r++) {
      for (int c = 1; c < 9; c++) {

        if (r > 1 && r < 8 && c > 1 && c < 8) {
          continue;
        }

        shrubs.add(new Shrub(c, r));
      }
    }

    sheeps.add(new ReverseSheep(5, 5, true));
  }
  else if (level == 8) {

    gridWidthInSquares = 10;
    gridHeightInSquares = 10;

    goalX = 9;
    goalY = 9;


    levelTitle = "Entourage";
    levelDescription = "Get the flock away.";

    sheeps.add(new Sheep(5, 5, false));

    sheeps.add(new Sheep(4, 4, true));
    sheeps.add(new Sheep(5, 4, true));
    sheeps.add(new Sheep(6, 4, true));
    sheeps.add(new Sheep(4, 5, true));
    sheeps.add(new Sheep(6, 5, true));
    sheeps.add(new Sheep(4, 6, true));
    sheeps.add(new Sheep(5, 6, true));
    sheeps.add(new Sheep(6, 6, true));

    shrubs.add(new Shrub(1, 1));
    shrubs.add(new Shrub(8, 8));
    shrubs.add(new Shrub(1, 8));
    shrubs.add(new Shrub(8, 1));
  }

  else if (level == 9) {

    gridWidthInSquares = 10;
    gridHeightInSquares = 10;

    goalX = 9;
    goalY = 9;


    levelTitle = "I Lava Ewe";
    levelDescription = "Lava burns up sheep!";
    sheeps.add(new Sheep(0, 0, false));


    sheeps.add(new Sheep(2, 2, true));
    sheeps.add(new Sheep(3, 3, true));
    sheeps.add(new Sheep(4, 4, true));
    sheeps.add(new Sheep(5, 5, true));
    
    shrubs.add(new Shrub(6, 6));
    
    for(int c = 2; c < 8; c++){
      lavas.add(new Lava(c, 6));
      lavas.add(new Lava(c, 7));
    }
    
     for(int r = 2; r < 6; r++){
      lavas.add(new Lava(6, r));
      lavas.add(new Lava(7, r));
    }

    
  }
  else if (level == 10) {

    gridWidthInSquares = 10;
    gridHeightInSquares = 10;

    goalX = 0;
    goalY = 9;


    levelTitle = "Baah Gates";
    levelDescription = "Stand on the button to open the gate.";



    sheeps.add(new Sheep(2, 3, false));


    sheeps.add(new Sheep(4, 0, true));

    for (int c = 0; c < 9; c++) {
      shrubs.add(new Shrub(c, 1));
    }

    Gate gate1 = new Gate(7, 3, 8, 3);
    Button button1 = new Button(9, 1, gate1);

    gates.add(gate1);
    buttons.add(button1);

    for (int c = 0; c < 9; c++) {
      shrubs.add(new Shrub(c, 1));
    }

    for (int c = 1; c < 10; c++) {
      shrubs.add(new Shrub(c, 2));
    }

    shrubs.add(new Shrub(8, 4));

    for (int c = 0; c < 9; c++) {
      shrubs.add(new Shrub(c, 5));
    }


    Gate gate2 = new Gate(5, 6, 6, 6);
    Button button2 = new Button(8, 7, gate2);

    gates.add(gate2);
    buttons.add(button2);

    sheeps.add(new Sheep(7, 6, true));

    for (int c = 1; c <= 5; c++) {
      gates.add(new Gate(c, 6, c, 7));
    }

    Gate gate3 = new Gate(8, 9, 9, 9);
    Button button3 = new Button(5, 7, gate3);

    gates.add(gate3);
    buttons.add(button3);

    shrubs.add(new Shrub(6, 7));
    shrubs.add(new Shrub(7, 7));
    shrubs.add(new Shrub(8, 6));

    for (int c = 0; c < 9; c++) {
      shrubs.add(new Shrub(c, 8));
    }
  }
  else if (level == 11) {

    gridWidthInSquares = 10;
    gridHeightInSquares = 10;

    goalX = 5;
    goalY = 9;


    levelTitle = "Spiral";
    levelDescription = "Warning: wool is very flammable.";


    sheeps.add(new Sheep(0, 0, false));
    sheeps.add(new ReverseSheep(7, 0, false));

    sheeps.add(new Sheep(1, 2, true));


    Gate gate = new Gate(5, 8, 5, 9);
    Button button = new Button(3, 6, gate);
    gates.add(gate);
    buttons.add(button);

    for (int r = 0; r < 10; r++) {
      lavas.add(new Lava(9, r));
    }


    for (int c = 0; c < 8; c++) {
      shrubs.add(new Shrub(c, 1));
    }

    for (int r = 3; r < 10; r++) {
      shrubs.add(new Shrub(0, r));
    }

    for (int c = 1; c < 7; c++) {
      shrubs.add(new Shrub(c, 3));
    }

    lavas.add(new Lava(7, 4));
    lavas.add(new Lava(7, 6));
    lavas.add(new Lava(7, 7));

      shrubs.add(new Shrub(7, 5));
    


    for (int c = 2; c < 7; c++) {
      shrubs.add(new Shrub(c, 7));
    }

    shrubs.add(new Shrub(2, 6));

    for (int c = 2; c < 6; c++) {
      shrubs.add(new Shrub(c, 5));
    }

    for (int c = 0; c < 9; c++) {
      if (c != 5) {
        shrubs.add(new Shrub(c, 9));
      }
    }
  }
  else if (level == 12) {

    gridWidthInSquares = 10;
    gridHeightInSquares = 10;

    goalX = 9;
    goalY = 5;


    levelTitle = "Cause and Effect";
    levelDescription = "Sorry I'm not sorry.";


    sheeps.add(new Sheep(1, 5, false));

    sheeps.add(new Sheep(1, 1, true));
    sheeps.add(new ReverseSheep(2, 8, true));
    sheeps.add(new Sheep(7, 7, true));
    sheeps.add(new Sheep(6, 1, true));

    //top row
    for (int c = 0; c < 10; c++) {
      if (c != 6) {
        shrubs.add(new Shrub(c, 0));
      }
    }

    //left column
    for (int r = 1; r < 7; r++) {
      if (r != 3) {
        shrubs.add(new Shrub(0, r));
      }
    }

    lavas.add(new Lava(6, 0));
    lavas.add(new Lava(0, 3));
    lavas.add(new Lava(2, 3));

    shrubs.add(new Shrub(2, 1));
    shrubs.add(new Shrub(2, 2));

    shrubs.add(new Shrub(7, 1));
    shrubs.add(new Shrub(7, 2));

    lavas.add(new Lava(1, 4));

    for (int c = 2; c < 10; c++) {
      shrubs.add(new Shrub(c, 4));
    }

    //middle row
    for (int c = 0; c < 10; c++) {
      shrubs.add(new Shrub(c, 6));
    }

    shrubs.add(new Shrub(3, 7));

    lavas.add(new Lava(0, 7));
    lavas.add(new Lava(1, 7));
    lavas.add(new Lava(2, 7));



    shrubs.add(new Shrub(5, 7));
    shrubs.add(new Shrub(6, 7));
    shrubs.add(new Shrub(5, 8));
    shrubs.add(new Shrub(6, 8));

    shrubs.add(new Shrub(8, 7));

    lavas.add(new Lava(0, 8));

    //bottom row
    for (int c = 0; c < 10; c++) {
      lavas.add(new Lava(c, 9));
    }





    Gate gate1 = new Gate(2, 5, 3, 5);
    Button button1 = new Button(4, 8, gate1);
    gates.add(gate1);
    buttons.add(button1);


    Gate gate2 = new Gate(4, 5, 5, 5);
    Button button2 = new Button(9, 7, gate2);
    gates.add(gate2);
    buttons.add(button2);


    Gate gate3 = new Gate(6, 5, 7, 5);
    Button button3 = new Button(6, 2, gate3);
    gates.add(gate3);
    buttons.add(button3);

    Gate gate4 = new Gate(8, 5, 9, 5);
    Button button4 = new Button(1, 3, gate4);
    gates.add(gate4);
    buttons.add(button4);
  }
}


void draw() 
{


  if (transitioning) {
    transition();
  }

  background(0);

  strokeWeight(1);


  if (showStartScreen) {  
    image(startScreen, 0, 0);
    return;
  }
  else if (showLevelScreen) {
    stroke(255);
    fill(255);
    textSize(36);

    float textWidth = textWidth("Level " + level);

    text("Level " + level, 250 - (textWidth/2), 180);
    textSize(24);

    textWidth = textWidth(levelTitle);
    text(levelTitle, 250 - (textWidth/2), 270);

    textSize(18);
    textWidth = textWidth(levelDescription);
    text(levelDescription, 250 - (textWidth/2), 350);

    stroke(100);
    fill(100);
    textWidth = textWidth("(click or space to begin)");
    text("(click or space to begin)", 250 - (textWidth/2), 480);

    return;
  }
  else if (showWinScreen) {

    stroke(255);
    fill(255);
    textSize(36);
    text("You Win!", 180, 220);
    textSize(22);
    text("(click or space for next level)", 100, 260);



    return;
  }
  else if (showFinalWinScreen) {
    image(finalWinScreen, 0, 0);
    return;
  }

  //workaround for loop() not behaving with music from the web
  /*if (music != null) {
    if (!music.isPlaying()) {
      music.rewind();
      music.play();
    }
  }*/



  background(color(0, 100, 0));

  image(grassBackground, 0, 0);

  int rectWidth = width/gridWidthInSquares;
  int rectHeight = height/gridHeightInSquares;


  image(flowersImage, goalX*rectWidth, goalY*rectHeight);


  for (int i = 0; i < shrubs.size(); i++) {
    ((Shrub)shrubs.get(i)).draw();
  }

  for (int i = 0; i < lavas.size(); i++) {
    ((Lava)lavas.get(i)).draw();
  }

  for (int i = 0; i < buttons.size(); i++) {
    ((Button)buttons.get(i)).draw();
  }





  for (int i = 0; i < sheeps.size(); i++) {

    if (!((Sheep)sheeps.get(i)).alive) {
      continue;
    }

    ((Sheep)sheeps.get(i)).draw();
  }





  fill(0);
  stroke(0);

  for (int i = 0; i < gridWidthInSquares; i++) {
    line(i*rectWidth, 0, i*rectWidth, height);
  }

  for (int i = 0; i < gridHeightInSquares; i++) {
    line(0, i*rectHeight, width, i*rectHeight);
  }

  for (int i = 0; i < gates.size(); i++) {
    ((Gate)gates.get(i)).draw();
  }


  noFill();
  strokeWeight(4);
  stroke(color(200, 100, 25));
  rect(0, 0, width-1, height-1);

  fill(color(15, 143, 56));
  strokeWeight(1);
  stroke(color(15, 143, 56));
  if (goalX == gridWidthInSquares-1) {
    rect((goalX+1) * rectWidth - 3, goalY*rectHeight, 3, rectHeight);
  }
  else{
      rect(goalX * rectWidth, (goalY+1)*rectHeight - 3, rectWidth, 3);
  }
}





void mousePressed() {
  nextScreen();
}

void nextScreen() {
  if (showStartScreen) {
    showStartScreen = false;
    showLevelScreen = true;
  }
  else if (showFinalWinScreen) {
    showFinalWinScreen = false;
    showStartScreen = true;
  }
  else if (showWinScreen) {
    showWinScreen = false;
    showLevelScreen = true;
  }
  else if (showLevelScreen) {
    showLevelScreen = false;
  }
}

void doAiMoves(int direction) {
  for (int i = 1; i < sheeps.size(); i++) {

    if (!((Sheep)sheeps.get(i)).alive) {
      continue;
    }

    if (!((Sheep)sheeps.get(i)).finishedTurn) {
      ((Sheep)sheeps.get(i)).move(direction);
    }
  }
}

void transition() {
  currentTransitionFrame ++;
  if (currentTransitionFrame >= transitionDuration) {

    doneTransitioning();
  }
}


void doneTransitioning() {

  transitioning = false;

  boolean somebodyDied = false;

  for (int s = 0; s < sheeps.size(); s++) {
    Sheep sheep = (Sheep)sheeps.get(s);
    if (sheep.alive) {
      for (int i = 0; i < lavas.size(); i++) {

        Lava lava = (Lava)lavas.get(i);
        if (sheep.x == lava.x && sheep.y == lava.y) {
          sheep.alive = false;
          System.out.println("Dead sheep.");
          somebodyDied = true;
          dieSound.trigger();
        }
      }
    }
  }

  if (somebodyDied) {
    dieSound.trigger();
  }



  if (!((Sheep)sheeps.get(0)).alive) {
    showLevelScreen = true;
    restart();
  }

  if (((Sheep)sheeps.get(0)).x == goalX && ((Sheep)sheeps.get(0)).y == goalY ) {
    if (goalX == gridWidthInSquares -1) {

      for (int s = 0; s < sheeps.size(); s++) {
        Sheep sheep = (Sheep)sheeps.get(s);
        if (sheep.alive) {
          sheep.prevX = sheep.x; 
          sheep.prevY = sheep.y;
        }
      }

      ((Sheep)sheeps.get(0)).x++;




      transitioning = true;
      currentTransitionFrame = 0;
    }
    else {

      for (int s = 0; s < sheeps.size(); s++) {
        Sheep sheep = (Sheep)sheeps.get(s);
        if (sheep.alive) {
          sheep.prevX = sheep.x; 
          sheep.prevY = sheep.y;
        }
      }

      ((Sheep)sheeps.get(0)).y++;


      transitioning = true;
      currentTransitionFrame = 0;
    }
    return;
  }


  if (((Sheep)sheeps.get(0)).x >= gridWidthInSquares || ((Sheep)sheeps.get(0)).y > gridHeightInSquares-1 ) {
    level++;

    if (level > 12) {
      showFinalWinScreen = true;
      level = 1;
    }
    else {
      showWinScreen = true;
    }

    restart();

    winSound.trigger();
  }




  for (int b = 0; b < buttons.size(); b++) {

    Button button = (Button)buttons.get(b);
    button.gate.open = false;

    for (int s = 0; s < sheeps.size(); s++) {
      Sheep sheep = (Sheep)sheeps.get(s);
      if (sheep.alive) {



        if (sheep.x == button.x && sheep.y == button.y) {
          button.gate.open = true;
        }
      }
    }
  }
}

void keyPressed() {

  if (transitioning) {
    return;
  }

  if (key == CODED) {


    boolean moved = false;

    for (int i = 0; i < sheeps.size(); i++) {
      Sheep otherSheep = (Sheep)sheeps.get(i); 
      otherSheep.finishedTurn = false;
    }

    if (keyCode == UP || keyCode == DOWN || keyCode == LEFT || keyCode == RIGHT) {

      double r = random(1);
      if (r < .25) {
        step1.trigger();
      }
      else if (r < .5) {
        step2.trigger();
      }
      else if (r < .75) {
        step3.trigger();
      }
      else {
        step4.trigger();
      }


      if (((Sheep)sheeps.get(0)).move(keyCode)) {
        moved = true;
      }
    }

    // if (moved) {
    doAiMoves(keyCode);
    // }


    transitioning = true;
    currentTransitionFrame = 0;
  }
  else if (keyCode == ' ') {
    nextScreen();
  }
  else if (key == 'r' || key == 'R') {
    showLevelScreen = true;
    dieSound.trigger();
    restart();
  }
  else if (key == 'm' || key == 'M') {

    if (winSound.isMuted()) {
      if (music != null) {
        music.unmute();
      }
      step1.unmute();
      step2.unmute();
      step3.unmute();
      step4.unmute();
      winSound.unmute();
      dieSound.unmute();
    }
    else {
      if (music != null) {
        music.mute();
      }
      step1.mute();
      step2.mute();
      step3.mute();
      step4.mute();
      winSound.mute();
      dieSound.mute();
    }
  }
}

boolean isBlocked(int fromX, int fromY, int toX, int toY) {

  if (toX == -1 || toX == gridWidthInSquares || toY == -1 || toY == gridHeightInSquares) {
    return true;
  }

  for (int i = 0; i < shrubs.size(); i++) {
    Shrub shrub = (Shrub)shrubs.get(i); 
    if (shrub.x == toX && shrub.y == toY) {
      return true;
    }
  }


  for (int i = 0; i < gates.size(); i++) {
    Gate gate = (Gate)gates.get(i);
    if (gate.open) {
      continue;
    }
    if (fromX == gate.x1 && fromY == gate.y1) {
      if (toX == gate.x2 && toY == gate.y2) {
        return true;
      }
    }
    else if (fromX == gate.x2 && fromY == gate.y2) {
      if (toX == gate.x1 && toY == gate.y1) {
        return true;
      }
    }
  }


  for (int i = 0; i < sheeps.size(); i++) {
    Sheep otherSheep = (Sheep)sheeps.get(i); 
    if (!otherSheep.alive) {
      continue;
    }
    if (otherSheep.x == toX && otherSheep.y == toY) {
      if (otherSheep.finishedTurn) {
        return true;
      }
      else {
        return !otherSheep.move(keyCode);
      }
    }
  }
  return false;
}


class Shrub {
  int x;
  int y;

  Shrub(int x, int y) {
    this.x = x;
    this.y = y;
  }

  void draw() {




    int rectWidth = width/gridWidthInSquares;
    int rectHeight = height/gridHeightInSquares;

    //ellipse(x*rectWidth + rectWidth/2, y*rectHeight + rectHeight/2, rectWidth/3, rectHeight/3);

    image(shrubImage, x*rectWidth, y*rectHeight, rectWidth, rectHeight);
  }
}

class Lava {
  int x;
  int y;

  Lava(int x, int y) {
    this.x = x;
    this.y = y;
  }

  void draw() {
    int rectWidth = width/gridWidthInSquares;
    int rectHeight = height/gridHeightInSquares;

    image(lavaImage, x*rectWidth, y*rectHeight, rectWidth, rectHeight);
  }
}


class Button {
  int x;
  int y;

  Gate gate;

  Button(int x, int y, Gate gate) {
    this.x = x;
    this.y = y;
    this.gate = gate;
  }

  void draw() {
    int rectWidth = width/gridWidthInSquares;
    int rectHeight = height/gridHeightInSquares;

    fill(color(100, 100, 100));
    stroke(color(100, 100, 100));

    rect(x*rectWidth, y*rectHeight, rectWidth, rectHeight);

    fill(color(0, 200, 0));
    stroke(color(0, 200, 0));

    ellipse(x*rectWidth +rectWidth/2, y*rectHeight + rectHeight/2, 10, 10);
  }
}


class Gate {
  int x1;
  int y1;
  int x2;
  int y2;

  int drawX;
  int drawY;
  int drawWidth;
  int drawHeight;

  boolean open = false;

  Gate(int x1, int y1, int x2, int y2) {
    this.x1 = x1;
    this.y1 = y1;
    this.x2 = x2;
    this.y2 = y2;


    int rectWidth = width/gridWidthInSquares;
    int rectHeight = height/gridHeightInSquares;

    if (x1 == x2) { //if up-and-down

      drawX = x1 * rectWidth;
      drawY = y2 * rectHeight - 10;
      drawWidth = rectWidth;
      drawHeight = 20;
    }
    else {  //if side-by-side
      drawX = x2 * rectWidth - 10;
      drawY = y1 * rectHeight;
      drawWidth = 20;
      drawHeight = rectHeight;
    }
  }

  void draw() {


    fill(color(200, 100, 25));
    stroke(color(225, 125, 50));


    if (!open) {
      rect(drawX, drawY, drawWidth, drawHeight);
    }
  }
}


class Sheep {
  int x;
  int y;
  boolean white;

  boolean alive = true;

  int prevX;
  int prevY;

  boolean finishedTurn = true;

  PImage mySheepImage = sheepImage;

  Sheep(int x, int y, boolean white) {
    this.x = x;
    this.y = y;
    this.white = white;
    if (!white) {
      mySheepImage = blackSheepImage;
    }
  }

  boolean goUp() { 
    if (isBlocked(x, y, x, y-1) ) {
      return false;
    }
    else {
      y = y-1;
      return true;
    }
  }

  boolean goDown() {
    if (isBlocked(x, y, x, y+1)) {
      return false;
    }
    else {
      y = y+1;
      return true;
    }
  }

  boolean goLeft() {
    if (isBlocked(x, y, x-1, y) ) {
      return false;
    }
    else {
      x = x-1;
      return true;
    }
  }

  boolean goRight() {
    if (isBlocked(x, y, x+1, y)) {
      return false;
    }
    else {
      x = x+1;
      return true;
    }
  }



  boolean move(int direction) {


    finishedTurn = true;

    boolean moved = false;

    prevX = x;
    prevY = y;

    if (direction == UP) {
      moved = goUp();
    }
    else if (direction == DOWN) {
      moved = goDown();
    }
    else if (direction == LEFT) {
      moved = goLeft();
    }
    else if (direction == RIGHT) {
      moved = goRight();
    }




    return moved;
  }

  void draw() {

    int rectWidth = width/gridWidthInSquares;
    int rectHeight = height/gridHeightInSquares;



    if (transitioning) {
      int xPerFrame = (rectWidth * (x - prevX))/transitionDuration;
      int yPerFrame = (rectHeight * (y - prevY))/transitionDuration;


      image(mySheepImage, prevX*rectWidth + (xPerFrame*currentTransitionFrame), prevY*rectHeight + (yPerFrame*currentTransitionFrame));

      return;
    }

    image(mySheepImage, x*rectWidth, y*rectHeight);
  }
}

class ReverseSheep extends Sheep {

  ReverseSheep(int x, int y, boolean white) {
    super(x, y, white);
    mySheepImage = reverseSheepImage;
  }

  boolean move(int direction) {

    finishedTurn = true;

    if (direction == UP) {
      return super.move(DOWN);
    }
    else if (direction == DOWN) {
      return super.move(UP);
    }
    else if (direction == LEFT) {
      return super.move(RIGHT);
    }
    else if (direction == RIGHT) {
      return super.move(LEFT);
    }
    return false;
  }
}

class LeftOnlySheep extends Sheep {

  LeftOnlySheep(int x, int y, boolean white) {
    super(x, y, white);
    mySheepImage = leftOnlySheepImage;
  }

  boolean move(int direction) {
    return super.move(LEFT);
  }
}

